"Program for 'string averaging' in Python3."

##  AVWORDS :
##  Program to test hybridization of strings/words.
##  Copyright 2012, R.S. Forsyth.
##  First version: 14/07/2010
##  Last revision: 24/12/2012


##  p3 libraries :
import  math, os, random, sys
import  time

##  local libraries :
sys.path.append(os.pardir+os.sep+"libs")

from globdefs import *
from glib import openfile
from glib import parafind, paraget
from glib import parashow
from glib import try2read


me = sys.argv[0]
print(me)  ##  progname


##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ("jobname","avwords"),  ##  name to link related files
    ("casefold",1,0,1),  ##  whether to fold upper & lower case
    ("voclists","lexicons.txt"),  ##  file containing path-specs of bilingual lexicons
    ("outgloss",''),  ##  o/p file for resultant glossary
    ("withkey",0,0,1),  ##  whether to include source term
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"avwords.txt"),  ##  listing file
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  globals :
Version = "3.0"
Feffmarx = ["\ufeff","\uffef"]  ##  to detect Windows faffing around with feff
Glostail = "_glos"


def main():
    ##  main line :
    
    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time()
    
    parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    na = len(args)
    print(progname,Version, dateline)
    print("no. of command-line args =", na)

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()
    
    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if not paradat.outpath.endswith(os.path.sep):
        paradat.outpath = paradat.outpath + os.path.sep
        ##  must have final separator character
        
    ##  read list of bilingual lexicons :
    filename = paradat.voclists
    lexfile = open(filename,'r',encoding="utf8")
    datlines = lexfile.readlines()
    lexfile.close()
    ##  expects each nonblank line of lexicon to have following format :
    ##  source-language-term Tabcharacter target-language-term
    ##  i.e. 2 items separated by a tab.

    ##  compile master dictionary of target-language equivalents :
    maintab = {}
    for datline in datlines:
        lexline = datline.strip()
        linesize = len(lexline)
        if linesize > 1:
            rawlines = try2read(lexline)
            print(len(rawlines),"lines in",lexline)
            for fm in Feffmarx:
                if len(rawlines) > 0 and rawlines[0].startswith(fm):
                    rawlines[0] = rawlines[0][1:]  ##  de-feffing
                    ##  print(rawlines[0])
            ##  process each line :
            for r in rawlines:
                if paradat.casefold:
                    r = r.lower()
                bits = r.strip().split(Htab)
                nb = len(bits)
                if nb != 2:  ##  problem
                    print("!!",bits)
                else:  ##  append tranlation of source term sw
                    sw, tw = bits
                    ##  remove extra blanks if present :
                    sw = sw.strip() ; tw = tw.strip()
                    tranlist = maintab.get(sw,[])
                    tranlist.append(tw)
                    maintab[sw] = tranlist

    print(len(maintab),"source-language terms.")
    keywords = sorted(maintab.keys())
    nk = len(keywords)
    if nk > 1:
        print(keywords[0],"....",keywords[-1])

    ##  write onto listing file :
    fo = open(paradat.outfile,'w',encoding="utf8")
    parashow(paradat,fo)

    ##  prepare o/p dump file :
    if paradat.outgloss:
        dumpname = paradat.outgloss
    else:
        dumpname = paradat.outpath + paradat.jobname + Glostail + ".dat"
    paradat.dumpfile = open(dumpname,'w',encoding="utf8")
    print(dateline,file=paradat.dumpfile)
    
    outlist = []
    outpairs = []
    ##  form 'average' of target terms for each source term :
    for kw in keywords:
        termlist = maintab[kw]
        if paradat.withkey > 0:  ##  include source item as well
            termlist.append(kw)
        print(kw,termlist,file=fo)
        gramtab = makegtab(termlist)
        termlens = [len(t) for t in termlist]
        ta = sum(termlens) / len(termlens)
        tm = midvalue(termlens)
        targsize = int(tm+0.5)
        ##  print(len(gramtab))
        ##  termean = avstring(termlist,targsize,0)
        ##  termean = stringav(termlist,1)
        fittest = 0 ; besterm = "zonQ"
        for t in termlist:
            tv = fitscore('<'+t+'>',gramtab)
            if tv > fittest:
                fittest = tv
                besterm = t
        termean = besterm  ##  analogous to median
        ev = fitscore('<' + termean + '>',gramtab)
        ##  print(termean,ev)
        tt = try2fix(termean,termlist,gramtab)
        tv = fitscore('<'+tt+'>',gramtab)
        print(tt,tv,file=fo)
        outlist.append((tv,tt))
        outpairs.append((kw,tt))

    ##  write o/p glossary :
    for st,tt in outpairs:
        print(st,Htab,tt,file=paradat.dumpfile)
    paradat.dumpfile.close()

    ##  write some evaluation info on listing file :
    print(NL,len(outlist),NL,file=fo)
    vtot = 0
    outlist.sort(reverse=True)
    for tv,tt in outlist:
        print(tt,tv,file=fo)
        vtot += tv
    print(vtot,file=fo)
    fo.close()
    
    print(":-)")
    print(time.asctime(time.localtime(time.time())))
    
    return




##  MIDVALUE :
def midvalue (xvec):
    "simple median."

    vecsize = len(xvec)
    if vecsize < 1:
        return  0.0  ##  None better ?
    if vecsize % 2 == 1:  ##  uneven
        m = vecsize // 2
        return  sorted(xvec)[m]
    else:  ##  even
        m = (vecsize-1)//2
        v1,v2 = sorted(xvec)[m:m+2]
        return  (v1+v2)*0.5


##  FITSCORE :
def fitscore (term,gramtab):
    "scores term by matches attested in gramtab, over length."

    termsize = len(term)
    gramtot = 0
    for p in range(termsize-1):
        c2 = term[p:p+2]  ##  character pair
        f = gramtab.get(c2,0)
        gramtot += f

    return  gramtot / termsize


##  MAKEGTAB :
def makegtab (termlist):
    "returns digram frequency table from termlist."

    gramtab = {}
    for term in termlist:
        t = '<' + term + '>'
        for j in range(len(t)-1):
            c2 = t[j:j+2]
            f = gramtab.get(c2,0)
            gramtab[c2] = f + 1

    return  gramtab
    

##  TRY2FIX :
def try2fix (newterm,termlist,gramtab):
    "tries to improve match of constructed term to termlist."

    ts = fitscore('<'+newterm+'>',gramtab)
    termsize = len(newterm)
    best = newterm
    ##  try swapping adjacent characters :
    for j in range(termsize-1):
        xx = newterm[j+1] + newterm[j]
        t = newterm[:j] + xx + newterm[j+2:]
        tv = fitscore('<'+t+'>',gramtab)
        if tv > ts:
            ts = tv ; best = t
            ##  print(best,tv)

    if termsize < 4:
        return  best  ##  don't try2 chop very short words
    ##  also try dropping characters :
    newbest = ''
    for j in range(termsize):
        newt = best[:j] + best[j+1:]
        tv = fitscore('<'+newt+'>',gramtab)
        if tv > ts:
            newbest = newt ; ts = tv
            ##  print(newbest,tv)

    if newbest:
        best = newbest

    return  best


##  AVSTRING :
def avstring (termlist,targsize,usemeans=0):
    "forms an 'average' of the strings (words) in termlist."

    combined = ''
    for term in termlist:
        combined = combined + term
    charset = set(combined)  ##  union of characters in input terms

    ##  tabulate frequency in each term :
    charfrex = {}
    for ch in charset:
        frex = []
        for term in termlist:
            cn = term.count(ch)
            frex.append(cn)
        charfrex[ch] = frex

    ##  determine precedence of characters :
    totfreq = 0
    freqlist = []
    for ch in charset:
        fx = sorted(charfrex[ch])
        ##  m = len(fx) // 2
        m = int(midvalue(fx)+0.5)
        a = sum(fx) / len(fx)
        av = (m*Gold+a)/(1+Gold)
        av = m
        ##  print(ch,charfrex[ch])
        nz = [f for f in fx if f > 0]
        totfreq += m
        freqlist.append((round(av,5),len(nz),ch))
    print(totfreq)
    freqlist.sort(reverse=True)
    print(freqlist)

    ##  then compute o/p frequencies :
    totfreq = 0
    charouf = {}
    for item in freqlist:
        nx = int(item[0]+0.5)
        ch = item[2]
        charouf[ch] = nx
        ##  print(ch,nx)
        totfreq += nx
        if totfreq >= targsize:
            break
    ##  print(charouf.keys())
    ##  compensate if not enough :
    if totfreq < targsize:
        print("----")
        for av,nz,ch in freqlist:
            if not (ch in charouf) or charouf[ch] <= 0:
                charouf[ch] = 1
                totfreq += 1
                if totfreq >= targsize:
                    break
    if totfreq > targsize:
        print("++",charouf)
    for item in charouf:
        if charouf[item] > 0:
            print(item,charouf[item])

    ##  next find relative positions of kept characters :
    charpos = {}
    for term in termlist:
        p = 0
        lent = len(term) - 1
        for ch in term:
            if ch not in charouf:
                continue
            if charouf[ch] > 0:
                locs = charpos.get(ch,[])
                locs.append(round(p/lent,5))
                charpos[ch] = locs
            p += 1
    print('#',len(charpos))
    ##  then compute mean positions in output term :
    tout = []
    for ch in charouf:
        nout = charouf[ch]
        if nout <= 0:  ##  not used
            continue
        locs = sorted(charpos[ch])
        if usemeans:
            blocavs = submeans(locs,nout)
            print("::",ch,locs)
            print(blocavs)
        else:  ##  use medians
            blocavs = submids(locs,nout)
            print("::",ch,locs)
            print(blocavs)
        ##  print("!!",ch,locs,blocavs)
        for p in blocavs:
            tout.append((p,ch))  ##  mean position(s)

    ##  construct o/p term :
    tout.sort()
    print(tout)  ##  testing
    outterm = ''
    for pair in tout:
        outterm = outterm + pair[1]
        
    return  outterm
            

##  STRINGAV :
def stringav (termlist,usemeans=0):
    "forms an 'average' of the strings (words) in termlist."

    combined = ''
    for term in termlist:
        combined = combined + term
    charset = set(combined)  ##  union of characters in input terms

    ##  tabulate frequency in each term :
    charfrex = {}
    for ch in charset:
        frex = []
        for term in termlist:
            cn = term.count(ch)
            frex.append(cn)
        charfrex[ch] = frex

    ##  determine frequency of chars in output :
    for ch in charset:
        fx = sorted(charfrex[ch])
        ##  m = len(fx) // 2
        ##  charfrex[ch] = fx[m]  ##  median as long as len(termlist) is odd
        charfrex[ch] = int(midvalue(fx)+0.5)

    ##  next find relative positions of kept characters :
    charpos = {}
    for term in termlist:
        p = 0
        lent = len(term) - 1
        for ch in term:
            if charfrex[ch] > 0:
                locs = charpos.get(ch,[])
                locs.append(round(p/lent,5))
                charpos[ch] = locs
            p += 1
    print('#',len(charpos))
    ##  then compute mean positions in output term :
    tout = []
    for ch in charfrex:
        nout = charfrex[ch]
        if nout <= 0:  ##  not used
            continue
        locs = sorted(charpos[ch])
        if usemeans:
            blocavs = submeans(locs,nout)
        else:  ##  use medians
            blocavs = submids(locs,nout)
        ##  print("!!",ch,locs,blocavs)
        for p in blocavs:
            tout.append((p,ch))  ##  mean position(s)

    ##  construct o/p term :
    tout.sort()
    print(tout)  ##  testing
    outterm = ''
    for pair in tout:
        outterm = outterm + pair[1]
        
    return  outterm


##  SUBMEANS :
def submeans (locs,blox):
    "returns list of blox means of sorted values in locs."

    blocsize = len(locs) / blox  ##  keep as floating-point
    meanlist = []
    for b in range(blox):
        p1 = math.trunc(b*blocsize)
        p2 = math.trunc((b+1)*blocsize+0.5)
        bloc = locs[p1:p2]
        av = round(sum(bloc) / len(bloc),4)
        meanlist.append(av)

    return  meanlist


##  SUBMIDS :
def submids (locs,blox):
    "returns list of blox medians of sorted values in locs."

    blocsize = len(locs) / blox  ##  keep as floating-point
    midlist = []
    for b in range(blox):
        p1 = math.trunc(b*blocsize)
        p2 = math.trunc((b+1)*blocsize+0.5)
        ##  bloc = sorted(locs[p1:p2])
        bloc = locs[p1:p2]
        ##  m = len(bloc) // 2
        ##  av = bloc[m]
        av = midvalue(bloc)
        midlist.append(round(av,4))

    return  midlist



if __name__ == "__main__":
    main()
