"Python3 program to make consensus vocabulary from LESU lists."

##  COREVOX :
##
##  Program to make core vocabulary by combining LESU lists :
##  Copyright 2012, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 22/12/2012
##  last revision : 24/12/2012


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)

import  math, os, string, sys
import  time


sys.path.append(os.pardir+os.sep+"libs")

##  local libraries :
from globdefs import *
from glib import add2dic, dicshow, numeric, numerize
from glib import openfile, paraget, parashow, showobj
from glib import termfrex, tokline, metaget, metachex
from glib import sibpath, toklines
from glib import parafind



##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ("jobname","corevox"),  ##  name to link related files
    ("casefold",1,0,1),  ##  whether to fold upper & lower case
    ("voclists","lesu.dat"),  ##  file containing path-specs of lesu files
    ("vocfile",''),  ##  o/p file for consensus list
    ("minfreq",2,1,),  ##  minimum freq 2b included in output term-list
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"corevox.txt"),  ##  listing file
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  Global vars :
Version = "1.1"
Mindocs = 2  ##  minimum number of lesu files
Voctail = "_lesu"
Outtail = "_vocs"
Feffmarx = ["\ufeff","\uffef"]  ##  to detect Windows faffing around with feff


##  Main program :
##
def main():
    "Main line."

    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time()
    
    parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()
    
    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if not paradat.outpath.endswith(os.path.sep):
        paradat.outpath = paradat.outpath + os.path.sep
        ##  must have final separator character
    ##  parashow(paradat)
    ##  tinf = Tokstuff(paraname)
    ##  paradat.tokinfo = tinf  ##  tokenization info passed thru

    ##  open input file :
    if paradat.voclists and os.path.isfile(paradat.voclists):
        paradat.voxname = paradat.voclists
    else:  ##  make a default name
        paradat.voxname = paradat.outpath + paradat.jobname + Voctail + ".dat"
    ##  paradat.gridfile = open(paradat.gridname,'r',encoding="utf-8")

    if not os.path.exists(paradat.outpath):
        os.makedirs(paradat.outpath)
    fi = open(paradat.voxname,'r',encoding="utf8")
        
    ##  write onto listing file :
    fo = open(paradat.outfile,'w',encoding="utf8")
    parashow(paradat,fo)

    ##  prepare o/p dump file :
    if paradat.vocfile:
        dumpname = paradat.vocfile
    else:
        dumpname = paradat.outpath + paradat.jobname + Outtail + ".dat"
    paradat.dumpfile = open(dumpname,'w',encoding="utf8")
    print(dateline,file=paradat.dumpfile)
    ##  print(paradat.progname,file=paradat.dumpfile)
    ##  print(paradat.id,file=paradat.dumpfile)

    ##  main process :
    maintab = {}
    vc = 0
    ##  read input lists :
    filelist = fi.readlines()
    for f in filelist:
        fn = f.strip()
        if os.path.isfile(fn):
            thislist = try2read(fn)  ##  assumes 1 item on each line (or zero)
        vset = set()  ##  set of items
        ##  append single copies of nonnull items :
        for item in thislist:
            it = item.strip()  ##  remove trailing nl et cetera
            if len(it) > 0:
                if paradat.casefold > 0:
                    it = it.lower()
                vset.add(it)
        ##  accumulate frex in maintab :
        for v in vset:
            freq = maintab.get(v,0)
            maintab[v] = freq + 1
        print(vc,fn)
        print(len(vset),len(maintab))
        vc += 1

    outlesu = tabsort(maintab,paradat)
    print()
    entries = len(outlesu)
    for j in range(10):
        print(j,outlesu[j])
    print("....")
    for j in range(entries-11,entries-1):
        print(j,outlesu[j])

    outlist = []
    for v in outlesu:
        print(v,file=fo)
        outlist.append(v[1])  ##  just keep word, without freq

    outlist.sort()
    for vo in outlist:
        print(vo,file=paradat.dumpfile)
        
    fo.close()
    paradat.dumpfile.close()
    
    ##  finale :
    print("\nOutput listing on :",paradat.outfile)
    print("Voclist dumped onto :",paradat.dumpfile.name,NL)
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,5),"seconds.")  ##  secs since kickoff
    
    return  0



##  TABSORT :
def tabsort (ftab,paradat):
    "sorts freq table, descending, and returns as list."

    fvec = []
    for k in ftab:
        f = ftab[k]
        if f >= paradat.minfreq:
            fvec.append((f,k))

    fvec.sort(reverse=True)

    return  fvec


##  GETFREX :
def getfrex (toklist):
    "makes frequency table for a single token list."

    freqtab = {}
    for t in toklist:
        f = freqtab.get(t,0)
        freqtab[t] = f + 1

    return  freqtab


##  MAKEFULL :
def makefull (prepath,filename):
    "joins prepath to filename, with separator if missing."

    if prepath.endswith(os.sep):
        fullname = prepath + filename
    else:  ##  insert separator
        fullname = prepath + os.sep + filename

    return  fullname


##  TRY2READ :
def try2read (f,codelist=["utf8","windows-1252","latin1"]):
    "tries more than once to read text file by guessing (?Windows) encodings."
    ##  utf8 is normally first choice.
    
    xxxx = None
    for ec in codelist:
        try:
            textfile = open(f,'r',encoding=ec)
            rawlines = textfile.readlines()
            textfile.close()
            return  rawlines
        except (UnicodeError,ValueError,IOError) as xxxx:
            rawlines = []

    ##  back-stop procedure is to replace error bytes (sort-of) :
    textfile = open(f,'r',encoding="utf8",errors="surrogateescape")
    rawlines = textfile.readlines()
    textfile.close()
    
    return  rawlines



        
##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
